@@ -21,6 +21,7 @@
 
 #include "config.h"
 #include <glib/gi18n-lib.h>
+#include <gio/gunixmounts.h>
 
 #include <string.h>
 #include <stdlib.h>
@@ -40,8 +41,6 @@
 #include "udiskslinuxblocklvm2.h"
 
 #include "udiskslvm2daemonutil.h"
-#include "udiskslvm2dbusutil.h"
-#include "udisks-lvm2-generated.h"
 #include "jobhelpers.h"
 
 /**
@@ -62,7 +61,7 @@ struct _UDisksLinuxVolumeGroupObject
 {
   UDisksObjectSkeleton parent_instance;
 
-  UDisksDaemon *daemon;
+  UDisksLinuxModuleLVM2 *module;
 
   gchar *name;
 
@@ -71,6 +70,8 @@ struct _UDisksLinuxVolumeGroupObject
   guint poll_timeout_id;
   gboolean poll_requested;
 
+  GUnixMountMonitor *mount_monitor;
+
   /* interface */
   UDisksVolumeGroup *iface_volume_group;
 };
@@ -83,15 +84,17 @@ struct _UDisksLinuxVolumeGroupObjectClass
 enum
 {
   PROP_0,
-  PROP_DAEMON,
+  PROP_MODULE,
   PROP_NAME,
 };
 
 G_DEFINE_TYPE (UDisksLinuxVolumeGroupObject, udisks_linux_volume_group_object, UDISKS_TYPE_OBJECT_SKELETON);
 
-static void etctabs_changed (UDisksFstabMonitor *monitor,
-                             UDisksFstabEntry   *entry,
-                             gpointer            user_data);
+static void fstab_changed (GUnixMountMonitor *monitor,
+                           gpointer           user_data);
+static void crypttab_changed (UDisksCrypttabMonitor  *monitor,
+                              UDisksCrypttabEntry    *entry,
+                              gpointer                user_data);
 
 typedef struct {
   BDLVMVGdata *vg_info;
@@ -102,8 +105,11 @@ static void
 udisks_linux_volume_group_object_finalize (GObject *_object)
 {
   UDisksLinuxVolumeGroupObject *object = UDISKS_LINUX_VOLUME_GROUP_OBJECT (_object);
+  UDisksDaemon *daemon;
 
-  /* note: we don't hold a ref to object->daemon */
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (object->module));
+
+  g_object_unref (object->module);
 
   if (object->iface_volume_group != NULL)
     g_object_unref (object->iface_volume_group);
@@ -111,12 +117,13 @@ udisks_linux_volume_group_object_finalize (GObject *_object)
   g_hash_table_unref (object->logical_volumes);
   g_free (object->name);
 
-  g_signal_handlers_disconnect_by_func (udisks_daemon_get_fstab_monitor (object->daemon),
-                                        G_CALLBACK (etctabs_changed),
+  g_signal_handlers_disconnect_by_func (object->mount_monitor,
+                                        G_CALLBACK (fstab_changed),
                                         object);
-  g_signal_handlers_disconnect_by_func (udisks_daemon_get_crypttab_monitor (object->daemon),
-                                        G_CALLBACK (etctabs_changed),
+  g_signal_handlers_disconnect_by_func (udisks_daemon_get_crypttab_monitor (daemon),
+                                        G_CALLBACK (crypttab_changed),
                                         object);
+  g_object_unref (object->mount_monitor);
 
   if (G_OBJECT_CLASS (udisks_linux_volume_group_object_parent_class)->finalize != NULL)
     G_OBJECT_CLASS (udisks_linux_volume_group_object_parent_class)->finalize (_object);
@@ -132,8 +139,8 @@ udisks_linux_volume_group_object_get_property (GObject    *__object,
 
   switch (prop_id)
     {
-    case PROP_DAEMON:
-      g_value_set_object (value, udisks_linux_volume_group_object_get_daemon (object));
+    case PROP_MODULE:
+      g_value_set_object (value, udisks_linux_volume_group_object_get_module (object));
       break;
 
     case PROP_NAME:
@@ -156,10 +163,9 @@ udisks_linux_volume_group_object_set_property (GObject      *__object,
 
   switch (prop_id)
     {
-    case PROP_DAEMON:
-      g_assert (object->daemon == NULL);
-      /* we don't take a reference to the daemon */
-      object->daemon = g_value_get_object (value);
+    case PROP_MODULE:
+      g_assert (object->module == NULL);
+      object->module = g_value_dup_object (value);
       break;
 
     case PROP_NAME:
@@ -186,11 +192,14 @@ static void
 udisks_linux_volume_group_object_constructed (GObject *_object)
 {
   UDisksLinuxVolumeGroupObject *object = UDISKS_LINUX_VOLUME_GROUP_OBJECT (_object);
+  UDisksDaemon *daemon;
   GString *s;
 
   if (G_OBJECT_CLASS (udisks_linux_volume_group_object_parent_class)->constructed != NULL)
     G_OBJECT_CLASS (udisks_linux_volume_group_object_parent_class)->constructed (_object);
 
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (object->module));
+
   object->logical_volumes = g_hash_table_new_full (g_str_hash,
                                                    g_str_equal,
                                                    g_free,
@@ -209,21 +218,18 @@ udisks_linux_volume_group_object_constructed (GObject *_object)
 
   /* Watch fstab and crypttab for changes.
    */
-  g_signal_connect (udisks_daemon_get_fstab_monitor (object->daemon),
+  object->mount_monitor = g_unix_mount_monitor_get ();
+  g_signal_connect (object->mount_monitor,
+                    "mountpoints-changed",
+                    G_CALLBACK (fstab_changed),
+                    object);
+  g_signal_connect (udisks_daemon_get_crypttab_monitor (daemon),
                     "entry-added",
-                    G_CALLBACK (etctabs_changed),
+                    G_CALLBACK (crypttab_changed),
                     object);
-  g_signal_connect (udisks_daemon_get_fstab_monitor (object->daemon),
+  g_signal_connect (udisks_daemon_get_crypttab_monitor (daemon),
                     "entry-removed",
-                    G_CALLBACK (etctabs_changed),
-                    object);
-  g_signal_connect (udisks_daemon_get_crypttab_monitor (object->daemon),
-                    "entry-added",
-                    G_CALLBACK (etctabs_changed),
-                    object);
-  g_signal_connect (udisks_daemon_get_crypttab_monitor (object->daemon),
-                    "entry-removed",
-                    G_CALLBACK (etctabs_changed),
+                    G_CALLBACK (crypttab_changed),
                     object);
 }
 
@@ -239,16 +245,16 @@ udisks_linux_volume_group_object_class_init (UDisksLinuxVolumeGroupObjectClass *
   gobject_class->get_property = udisks_linux_volume_group_object_get_property;
 
   /**
-   * UDisksLinuxVolumeGroupObject:daemon:
+   * UDisksLinuxVolumeGroupObject:module:
    *
-   * The #UDisksDaemon the object is for.
+   * The #UDisksLinuxModuleLVM2 the object is for.
    */
   g_object_class_install_property (gobject_class,
-                                   PROP_DAEMON,
-                                   g_param_spec_object ("daemon",
-                                                        "Daemon",
-                                                        "The daemon the object is for",
-                                                        UDISKS_TYPE_DAEMON,
+                                   PROP_MODULE,
+                                   g_param_spec_object ("module",
+                                                        "Module",
+                                                        "The module the object is for",
+                                                        UDISKS_TYPE_LINUX_MODULE_LVM2,
                                                         G_PARAM_READABLE |
                                                         G_PARAM_WRITABLE |
                                                         G_PARAM_CONSTRUCT_ONLY |
@@ -272,7 +278,7 @@ udisks_linux_volume_group_object_class_init (UDisksLinuxVolumeGroupObjectClass *
 
 /**
  * udisks_linux_volume_group_object_new:
- * @daemon: A #UDisksDaemon.
+ * @module: A #UDisksLinuxModuleLVM2.
  * @name: The name of the volume group.
  *
  * Create a new VolumeGroup object.
@@ -280,30 +286,30 @@ udisks_linux_volume_group_object_class_init (UDisksLinuxVolumeGroupObjectClass *
  * Returns: A #UDisksLinuxVolumeGroupObject object. Free with g_object_unref().
  */
 UDisksLinuxVolumeGroupObject *
-udisks_linux_volume_group_object_new (UDisksDaemon *daemon,
-                                      const gchar  *name)
+udisks_linux_volume_group_object_new (UDisksLinuxModuleLVM2 *module,
+                                      const gchar           *name)
 {
-  g_return_val_if_fail (UDISKS_IS_DAEMON (daemon), NULL);
+  g_return_val_if_fail (UDISKS_IS_LINUX_MODULE_LVM2 (module), NULL);
   g_return_val_if_fail (name != NULL, NULL);
   return UDISKS_LINUX_VOLUME_GROUP_OBJECT (g_object_new (UDISKS_TYPE_LINUX_VOLUME_GROUP_OBJECT,
-                                                         "daemon", daemon,
+                                                         "module", module,
                                                          "name", name,
                                                          NULL));
 }
 
 /**
- * udisks_linux_volume_group_object_get_daemon:
+ * udisks_linux_volume_group_object_get_module:
  * @object: A #UDisksLinuxVolumeGroupObject.
  *
- * Gets the daemon used by @object.
+ * Gets the module used by @object.
  *
- * Returns: A #UDisksDaemon. Do not free, the object is owned by @object.
+ * Returns: A #UDisksLinuxModuleLVM2. Do not free, the object is owned by @object.
  */
-UDisksDaemon *
-udisks_linux_volume_group_object_get_daemon (UDisksLinuxVolumeGroupObject *object)
+UDisksLinuxModuleLVM2 *
+udisks_linux_volume_group_object_get_module (UDisksLinuxVolumeGroupObject *object)
 {
   g_return_val_if_fail (UDISKS_IS_LINUX_VOLUME_GROUP_OBJECT (object), NULL);
-  return object->daemon;
+  return object->module;
 }
 
 static void
@@ -322,9 +328,16 @@ update_etctabs (UDisksLinuxVolumeGroupObject *object)
 }
 
 static void
-etctabs_changed (UDisksFstabMonitor *monitor,
-                 UDisksFstabEntry   *entry,
-                 gpointer            user_data)
+fstab_changed (GUnixMountMonitor *monitor,
+               gpointer           user_data)
+{
+  update_etctabs (UDISKS_LINUX_VOLUME_GROUP_OBJECT (user_data));
+}
+
+static void
+crypttab_changed (UDisksCrypttabMonitor  *monitor,
+                  UDisksCrypttabEntry    *entry,
+                  gpointer                user_data)
 {
   update_etctabs (UDISKS_LINUX_VOLUME_GROUP_OBJECT (user_data));
 }
@@ -336,25 +349,27 @@ lv_is_pvmove_volume (const gchar *name)
 }
 
 static void
-update_progress_for_device (UDisksDaemon *daemon,
-                            const gchar  *operation,
-                            const gchar  *dev,
-                            double        progress)
+update_progress_for_device (UDisksLinuxVolumeGroupObject *object,
+                            const gchar                  *operation,
+                            const gchar                  *dev,
+                            double                        progress)
 {
   GDBusObjectManager *object_manager;
+  UDisksDaemon *daemon;
   GList *objects, *l;
 
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (object->module));
   object_manager = G_DBUS_OBJECT_MANAGER (udisks_daemon_get_object_manager (daemon));
   objects = g_dbus_object_manager_get_objects (object_manager);
 
   for (l = objects; l; l = l->next)
     {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
+      UDisksObject *o = UDISKS_OBJECT (l->data);
       UDisksJob *job;
       const gchar *const *job_objects;
       int i;
 
-      job = udisks_object_peek_job (object);
+      job = udisks_object_peek_job (o);
       if (job == NULL)
         continue;
 
@@ -391,16 +406,16 @@ update_progress_for_device (UDisksDaemon *daemon,
 }
 
 static void
-update_operations (UDisksDaemon *daemon,
-                   const gchar  *lv_name,
-                   BDLVMLVdata  *lv_info,
-                   gboolean     *needs_polling_ret)
+update_operations (UDisksLinuxVolumeGroupObject *object,
+                   const gchar                  *lv_name,
+                   BDLVMLVdata                  *lv_info,
+                   gboolean                     *needs_polling_ret)
 {
   if (lv_is_pvmove_volume (lv_name))
     {
       if (lv_info->move_pv && lv_info->copy_percent)
         {
-          update_progress_for_device (daemon,
+          update_progress_for_device (object,
                                       "lvm-vg-empty-device",
                                       lv_info->move_pv,
                                       lv_info->copy_percent/100.0);
@@ -411,7 +426,7 @@ update_operations (UDisksDaemon *daemon,
 
 static void
 block_object_update_lvm_iface (UDisksLinuxBlockObject *object,
-                               const gchar *lv_obj_path)
+                               const gchar            *lv_obj_path)
 {
   UDisksBlockLVM2 *iface_block_lvm2;
 
@@ -427,6 +442,7 @@ block_object_update_lvm_iface (UDisksLinuxBlockObject *object,
 
   udisks_linux_block_lvm2_update (UDISKS_LINUX_BLOCK_LVM2 (iface_block_lvm2), object);
   udisks_block_lvm2_set_logical_volume (iface_block_lvm2, lv_obj_path);
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (iface_block_lvm2));
 }
 
 static void
@@ -441,6 +457,7 @@ lv_object_update_block_path (UDisksLinuxBlockObject       *block_object,
     {
       block_objpath = g_dbus_object_get_object_path (G_DBUS_OBJECT (block_object));
       udisks_logical_volume_set_block_device (UDISKS_LOGICAL_VOLUME (lv), block_objpath);
+      g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (lv));
     }
 }
 
@@ -457,7 +474,7 @@ update_block (UDisksLinuxBlockObject       *block_object,
   if (block == NULL)
     return;
 
-  // XXX - move this elsewhere?
+  /* XXX - move this elsewhere? */
   {
     UDisksLinuxDevice *device;
     UDisksLinuxLogicalVolumeObject *lv_object;
@@ -582,7 +599,7 @@ update_vg (GObject      *source_obj,
       return;
     }
 
-  daemon = udisks_linux_volume_group_object_get_daemon (object);
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (object->module));
   manager = udisks_daemon_get_object_manager (daemon);
 
   udisks_linux_volume_group_update (UDISKS_LINUX_VOLUME_GROUP (object->iface_volume_group), vg_info, &needs_polling);
@@ -598,8 +615,9 @@ update_vg (GObject      *source_obj,
       BDLVMLVdata *lv_info = *lvs_p;
       const gchar *lv_name = lv_info->lv_name;
       BDLVMLVdata *meta_lv_info = NULL;
+      BDLVMVDOPooldata *vdo_info = NULL;
 
-      update_operations (daemon, lv_name, lv_info, &needs_polling);
+      update_operations (object, lv_name, lv_info, &needs_polling);
 
       if (udisks_daemon_util_lvm2_name_is_reserved (lv_name))
         continue;
@@ -610,17 +628,29 @@ update_vg (GObject      *source_obj,
           if (cmp_int_lv_name ((*lvs_p2)->lv_name, lv_info->metadata_lv))
             meta_lv_info = *lvs_p2;
 
+      if (lv_info->pool_lv && g_strcmp0 (lv_info->segtype, "vdo") == 0)
+        {
+          vdo_info = bd_lvm_vdo_info (lv_info->vg_name, lv_info->pool_lv, &error);
+          if (!vdo_info)
+            {
+              udisks_warning ("Failed to get information about VDO volume %s: %s",
+                              lv_info->lv_name, error->message);
+              g_clear_error (&error);
+            }
+        }
+
+
       volume = g_hash_table_lookup (object->logical_volumes, lv_name);
       if (volume == NULL)
         {
-          volume = udisks_linux_logical_volume_object_new (daemon, object, lv_name);
-          udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, &needs_polling);
+          volume = udisks_linux_logical_volume_object_new (object->module, object, lv_name);
+          udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, vdo_info, &needs_polling);
           udisks_linux_logical_volume_object_update_etctabs (volume);
           g_dbus_object_manager_server_export_uniquely (manager, G_DBUS_OBJECT_SKELETON (volume));
           g_hash_table_insert (object->logical_volumes, g_strdup (lv_name), volume);
         }
       else
-        udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, &needs_polling);
+        udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, vdo_info, &needs_polling);
 
       g_hash_table_insert (new_lvs, (gchar *)lv_name, volume);
     }
@@ -667,6 +697,7 @@ update_vg (GObject      *source_obj,
   bd_lvm_vgdata_free (vg_info);
   lv_list_free (lvs);
 
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (object->iface_volume_group));
   g_object_unref (object);
 }
 
@@ -695,7 +726,6 @@ poll_vg_update (GObject      *source_obj,
                 GAsyncResult *result,
                 gpointer      user_data)
 {
-  UDisksDaemon *daemon;
   gboolean needs_polling;
   GError *error = NULL;
   UDisksLinuxVolumeGroupObject *object = UDISKS_LINUX_VOLUME_GROUP_OBJECT (source_obj);
@@ -729,8 +759,6 @@ poll_vg_update (GObject      *source_obj,
       return;
     }
 
-  daemon = udisks_linux_volume_group_object_get_daemon (object);
-
   /* XXX: we used to do this, but it seems to be pointless (how could a VG change without emitting a uevent on the PVs?) */
   /* udisks_linux_volume_group_update (UDISKS_LINUX_VOLUME_GROUP (object->iface_volume_group), info, &needs_polling); */
 
@@ -740,6 +768,7 @@ poll_vg_update (GObject      *source_obj,
       BDLVMLVdata *lv_info = *lvs_p;
       BDLVMLVdata *meta_lv_info = NULL;
       const gchar *lv_name = lv_info->lv_name;
+      BDLVMVDOPooldata *vdo_info = NULL;
 
       if (lv_info->metadata_lv && *(lv_info->metadata_lv) != '\0')
         /* this is not cheap to do, but not many LVs have a metadata LV */
@@ -747,10 +776,21 @@ poll_vg_update (GObject      *source_obj,
           if (cmp_int_lv_name ((*lvs_np)->lv_name, lv_info->metadata_lv))
             meta_lv_info = *lvs_np;
 
-      update_operations (daemon, lv_name, lv_info, &needs_polling);
+      if (lv_info->pool_lv && g_strcmp0 (lv_info->segtype, "vdo") == 0)
+        {
+          vdo_info = bd_lvm_vdo_info (lv_info->vg_name, lv_info->pool_lv, &error);
+          if (!vdo_info)
+            {
+              udisks_warning ("Failed to get information about VDO volume %s: %s",
+                              lv_info->lv_name, error->message);
+              g_clear_error (&error);
+            }
+        }
+
+      update_operations (object, lv_name, lv_info, &needs_polling);
       volume = g_hash_table_lookup (object->logical_volumes, lv_name);
       if (volume)
-        udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, &needs_polling);
+        udisks_linux_logical_volume_object_update (volume, lv_info, meta_lv_info, vdo_info, &needs_polling);
     }
 
   lv_list_free (lvs);
@@ -820,14 +860,17 @@ udisks_linux_volume_group_object_poll (UDisksLinuxVolumeGroupObject *object)
 void
 udisks_linux_volume_group_object_destroy (UDisksLinuxVolumeGroupObject *object)
 {
+  UDisksDaemon *daemon;
   GHashTableIter volume_iter;
   gpointer key, value;
 
+  daemon = udisks_module_get_daemon (UDISKS_MODULE (object->module));
+
   g_hash_table_iter_init (&volume_iter, object->logical_volumes);
   while (g_hash_table_iter_next (&volume_iter, &key, &value))
     {
       UDisksLinuxLogicalVolumeObject *volume = value;
-      g_dbus_object_manager_server_unexport (udisks_daemon_get_object_manager (object->daemon),
+      g_dbus_object_manager_server_unexport (udisks_daemon_get_object_manager (daemon),
                                              g_dbus_object_get_object_path (G_DBUS_OBJECT (volume)));
     }
 }
