@@ -50,6 +50,7 @@
 #include "udisksdaemon.h"
 #include "udisksdaemonutil.h"
 #include "udiskslinuxdevice.h"
+#include "udisksconfigmanager.h"
 
 /**
  * SECTION:udiskslinuxdrive
@@ -228,27 +229,26 @@ static const VariantKeyfileMapping drive_configuration_mapping[5] = {
 /* ---------------------------------------------------------------------------------------------------- */
 
 static gchar *
-configuration_get_path (UDisksLinuxDrive *drive)
+configuration_get_path (UDisksLinuxDrive *drive,
+                        UDisksDaemon     *daemon)
 {
+  UDisksConfigManager *config_manager;
   const gchar *id;
-  gchar *path = NULL;
+  gchar *id_config_file;
+  gchar *path;
+
+  config_manager = udisks_daemon_get_config_manager (daemon);
 
   id = udisks_drive_get_id (UDISKS_DRIVE (drive));
   if (id == NULL || strlen (id) == 0)
-    goto out;
+    return NULL;
 
-  /* if prefix is specified directories may not exist */
-  if (!g_file_test (PACKAGE_SYSCONF_DIR "/udisks2", G_FILE_TEST_IS_DIR))
-    {
-      if (g_mkdir_with_parents (PACKAGE_SYSCONF_DIR "/udisks2", 0700) != 0)
-        {
-          udisks_critical ("Error creating directory %s: %m", PACKAGE_SYSCONF_DIR "/udisks2");
-        }
-    }
+  id_config_file = g_strdup_printf ("%s.conf", id);
+  path = g_build_filename (udisks_config_manager_get_config_dir (config_manager),
+                           id_config_file,
+                           NULL);
+  g_free (id_config_file);
 
-  path = g_strdup_printf (PACKAGE_SYSCONF_DIR "/udisks2/%s.conf", id);
-
- out:
   return path;
 }
 
@@ -257,6 +257,7 @@ static gboolean
 update_configuration (UDisksLinuxDrive       *drive,
                       UDisksLinuxDriveObject *object)
 {
+  UDisksDaemon *daemon;
   GKeyFile *key_file = NULL;
   gboolean ret = FALSE;
   gchar *path = NULL;
@@ -266,7 +267,9 @@ update_configuration (UDisksLinuxDrive       *drive,
   GVariant *old_value;
   guint n;
 
-  path = configuration_get_path (drive);
+  daemon = udisks_linux_drive_object_get_daemon (object);
+
+  path = configuration_get_path (drive, daemon);
   if (path == NULL)
     goto out;
 
@@ -740,7 +743,8 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
       if (model != NULL)
         {
           gchar *s;
-          s = udisks_decode_udev_string (model);
+          s = udisks_decode_udev_string (model,
+                                         g_udev_device_get_property (device->udev_device, "ID_MODEL"));
           g_strstrip (s);
           udisks_drive_set_model (iface, s);
           g_free (s);
@@ -759,7 +763,8 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
       if (vendor != NULL)
         {
           gchar *s;
-          s = udisks_decode_udev_string (vendor);
+          s = udisks_decode_udev_string (vendor,
+                                         g_udev_device_get_property (device->udev_device, "ID_VENDOR"));
           g_strstrip (s);
           udisks_drive_set_vendor (iface, s);
           g_free (s);
@@ -769,7 +774,8 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
       if (model != NULL)
         {
           gchar *s;
-          s = udisks_decode_udev_string (model);
+          s = udisks_decode_udev_string (model,
+                                         g_udev_device_get_property (device->udev_device, "ID_MODEL"));
           g_strstrip (s);
           udisks_drive_set_model (iface, s);
           g_free (s);
@@ -801,7 +807,8 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
       if (vendor != NULL)
         {
           gchar *s;
-          s = udisks_decode_udev_string (vendor);
+          s = udisks_decode_udev_string (vendor,
+                                         g_udev_device_get_property (device->udev_device, "ID_VENDOR"));
           g_strstrip (s);
           udisks_drive_set_vendor (iface, s);
           g_free (s);
@@ -830,7 +837,8 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
       if (model != NULL)
         {
           gchar *s;
-          s = udisks_decode_udev_string (model);
+          s = udisks_decode_udev_string (model,
+                                         g_udev_device_get_property (device->udev_device, "ID_MODEL"));
           g_strstrip (s);
           udisks_drive_set_model (iface, s);
           g_free (s);
@@ -942,6 +950,7 @@ udisks_linux_drive_update (UDisksLinuxDrive       *drive,
   ret = update_configuration (drive, object);
 
  out:
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (drive));
   if (device != NULL)
     g_clear_object (&device);
 
@@ -1112,7 +1121,7 @@ handle_set_configuration (UDisksDrive           *_drive,
                                                     invocation))
     goto out;
 
-  path = configuration_get_path (drive);
+  path = configuration_get_path (drive, daemon);
   if (path == NULL)
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
