@@ -50,6 +50,7 @@
 #include "udisksthreadedjob.h"
 #include "udisksata.h"
 #include "udiskslinuxdevice.h"
+#include "udisksconfigmanager.h"
 
 /**
  * SECTION:udiskslinuxdriveata
@@ -338,6 +339,8 @@ udisks_linux_drive_ata_update (UDisksLinuxDriveAta    *drive,
   update_security (drive, device);
 
  out:
+  /* ensure property changes are sent before the method return */
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (drive));
   if (device != NULL)
     g_object_unref (device);
 
@@ -680,6 +683,9 @@ udisks_linux_drive_ata_refresh_smart_sync (UDisksLinuxDriveAta  *drive,
   /* update stats again to account for the IO we just did to read the SMART info */
   update_io_stats (drive, device);
 
+  /* ensure property changes are sent before the method return */
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (drive));
+
  out:
   g_clear_object (&device);
   if (d != NULL)
@@ -1622,15 +1628,17 @@ static gpointer
 apply_configuration_thread_func (gpointer user_data)
 {
   ApplyConfData *data = user_data;
+  UDisksDaemon *daemon;
   const gchar *device_file = NULL;
   gint fd = -1;
   GError *error = NULL;
 
+  daemon = udisks_linux_drive_object_get_daemon (data->object);
   device_file = g_udev_device_get_device_file (data->device->udev_device);
 
-  udisks_notice ("Applying configuration from %s/udisks2/%s.conf to %s",
-                 PACKAGE_SYSCONF_DIR, udisks_drive_get_id (data->drive), device_file);
-
+  udisks_notice ("Applying configuration from %s/%s.conf to %s",
+                 udisks_config_manager_get_config_dir (udisks_daemon_get_config_manager (daemon)),
+                 udisks_drive_get_id (data->drive), device_file);
 
   /* Use O_RDRW instead of O_RDONLY to force a 'change' uevent so properties are updated */
   fd = open (device_file, O_RDWR|O_NONBLOCK);
@@ -1640,33 +1648,6 @@ apply_configuration_thread_func (gpointer user_data)
       goto out;
     }
 
-  if (data->ata_pm_standby != -1)
-    {
-      /* ATA8: 7.18 IDLE - E3h, Non-Data */
-      UDisksAtaCommandInput input = {.command = 0xe3, .count = data->ata_pm_standby};
-      UDisksAtaCommandOutput output = {0};
-      if (!udisks_ata_send_command_sync (fd,
-                                         -1,
-                                         UDISKS_ATA_COMMAND_PROTOCOL_NONE,
-                                         &input,
-                                         &output,
-                                         &error))
-        {
-          udisks_critical ("Error sending ATA command IDLE (timeout=%d) to %s: %s (%s, %d)",
-                        data->ata_pm_standby, device_file,
-                        error->message, g_quark_to_string (error->domain), error->code);
-          g_clear_error (&error);
-        }
-      else
-        {
-          gchar *pretty = ata_pm_standby_to_string (data->ata_pm_standby);
-          udisks_notice ("Set standby timer to %s (value %d) on %s [%s]",
-                         pretty, data->ata_pm_standby, device_file, udisks_drive_get_id (data->drive));
-          g_free (pretty);
-          data->ata->standby_enabled = data->ata_pm_standby != 0;
-        }
-    }
-
   if (data->ata_apm_level != -1)
     {
       /* ATA8: 7.48 SET FEATURES - EFh, Non-Data
@@ -1698,6 +1679,33 @@ apply_configuration_thread_func (gpointer user_data)
         }
     }
 
+  if (data->ata_pm_standby != -1)
+    {
+      /* ATA8: 7.18 IDLE - E3h, Non-Data */
+      UDisksAtaCommandInput input = {.command = 0xe3, .count = data->ata_pm_standby};
+      UDisksAtaCommandOutput output = {0};
+      if (!udisks_ata_send_command_sync (fd,
+                                         -1,
+                                         UDISKS_ATA_COMMAND_PROTOCOL_NONE,
+                                         &input,
+                                         &output,
+                                         &error))
+        {
+          udisks_critical ("Error sending ATA command IDLE (timeout=%d) to %s: %s (%s, %d)",
+                        data->ata_pm_standby, device_file,
+                        error->message, g_quark_to_string (error->domain), error->code);
+          g_clear_error (&error);
+        }
+      else
+        {
+          gchar *pretty = ata_pm_standby_to_string (data->ata_pm_standby);
+          udisks_notice ("Set standby timer to %s (value %d) on %s [%s]",
+                         pretty, data->ata_pm_standby, device_file, udisks_drive_get_id (data->drive));
+          g_free (pretty);
+          data->ata->standby_enabled = data->ata_pm_standby != 0;
+        }
+    }
+
   if (data->ata_aam_level != -1)
     {
       /* ATA8: 7.48 SET FEATURES - EFh, Non-Data
@@ -2308,6 +2316,8 @@ handle_security_erase_unit (UDisksDriveAta        *_drive,
     }
 
   udisks_linux_block_object_reread_partition_table (UDISKS_LINUX_BLOCK_OBJECT (block_object));
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (block_object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
 
  out:
   g_clear_object (&block_object);
