@@ -41,7 +41,6 @@
 #include "udiskslinuxdevice.h"
 #include "udiskslinuxblock.h"
 #include "udisksfstabentry.h"
-#include "udisksfstabmonitor.h"
 #include "udiskscrypttabentry.h"
 #include "udiskscrypttabmonitor.h"
 #include "udisksspawnedjob.h"
@@ -234,9 +233,12 @@ udisks_linux_encrypted_update (UDisksLinuxEncrypted   *encrypted,
         udisks_block_set_id_type (block, "crypto_TCRYPT");
     }
 
-  update_metadata_size (encrypted, object);
+  if (udisks_linux_block_is_luks (block))
+    update_metadata_size (encrypted, object);
 
   udisks_linux_block_encrypted_unlock (block);
+
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (encrypted));
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -332,7 +334,7 @@ handle_unlock (UDisksEncrypted        *encrypted,
   UDisksObject *object = NULL;
   UDisksBlock *block;
   UDisksDaemon *daemon;
-  UDisksState *state;
+  UDisksState *state = NULL;
   gchar *name = NULL;
   UDisksObject *cleartext_object = NULL;
   UDisksBlock *cleartext_block;
@@ -357,8 +359,10 @@ handle_unlock (UDisksEncrypted        *encrypted,
   const gchar *keyfiles[MAX_TCRYPT_KEYFILES] = {};
   CryptoJobData data;
   gboolean is_luks;
+  gboolean is_bitlk;
   gboolean handle_as_tcrypt;
   void *open_func;
+  const gchar *uuid = NULL;
 
   object = udisks_daemon_util_dup_object (encrypted, &error);
   if (object == NULL)
@@ -371,8 +375,12 @@ handle_unlock (UDisksEncrypted        *encrypted,
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
   state = udisks_daemon_get_state (daemon);
   is_luks = udisks_linux_block_is_luks (block);
+  is_bitlk = udisks_linux_block_is_bitlk (block);
   handle_as_tcrypt = udisks_linux_block_is_tcrypt (block) || udisks_linux_block_is_unknown_crypto (block);
 
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
   /* get TCRYPT options */
   if (handle_as_tcrypt)
     {
@@ -403,12 +411,12 @@ handle_unlock (UDisksEncrypted        *encrypted,
    */
 
   /* Fail if the device is not a LUKS or possible TCRYPT device */
-  if (!(is_luks || handle_as_tcrypt))
+  if (!(is_luks || is_bitlk || handle_as_tcrypt))
     {
       g_dbus_method_invocation_return_error (invocation,
                                              UDISKS_ERROR,
                                              UDISKS_ERROR_FAILED,
-                                             "Device %s does not appear to be a LUKS or TCRYPT device",
+                                             "Device %s does not appear to be a LUKS, BITLK or TCRYPT device",
                                              udisks_block_get_device (block));
       goto out;
     }
@@ -516,6 +524,14 @@ handle_unlock (UDisksEncrypted        *encrypted,
   else {
     if (is_luks)
       name = g_strdup_printf ("luks-%s", udisks_block_get_id_uuid (block));
+    else if (is_bitlk)
+      {
+        uuid = udisks_block_get_id_uuid (block);
+        if (uuid && g_strcmp0 (uuid, "") != 0)
+          name = g_strdup_printf ("bitlk-%s", uuid);
+        else
+          name = g_strdup_printf ("bitlk-%" G_GUINT64_FORMAT, udisks_block_get_device_number (block));
+      }
     else
       /* TCRYPT devices don't have a UUID, so we use the device number instead */
       name = g_strdup_printf ("tcrypt-%" G_GUINT64_FORMAT, udisks_block_get_device_number (block));
@@ -529,12 +545,15 @@ handle_unlock (UDisksEncrypted        *encrypted,
    * update triggered by the unlock. */
   if (is_luks)
     udisks_encrypted_set_hint_encryption_type (encrypted, "LUKS");
+  else if (is_bitlk)
+    udisks_encrypted_set_hint_encryption_type (encrypted, "BITLK");
   else
     udisks_encrypted_set_hint_encryption_type (encrypted, "TCRYPT");
 
   device = udisks_block_dup_device (block);
 
-  /* TODO: support reading a 'readonly' option from @options */
+  /* unlock as read-only if specified in @options or if the device itself is read-only */
+  g_variant_lookup (options, "read-only", "b", &read_only);
   if (udisks_block_get_read_only (block))
     read_only = TRUE;
 
@@ -549,6 +568,8 @@ handle_unlock (UDisksEncrypted        *encrypted,
 
   if (is_luks)
     open_func = luks_open_job_func;
+  else if (is_bitlk)
+    open_func = bitlk_open_job_func;
   else
     open_func = tcrypt_open_job_func;
 
@@ -586,7 +607,7 @@ handle_unlock (UDisksEncrypted        *encrypted,
                                                          wait_for_cleartext_object,
                                                          g_strdup (g_dbus_object_get_object_path (G_DBUS_OBJECT (object))),
                                                          g_free,
-                                                         20, /* timeout_seconds */
+                                                         UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                          &error);
   if (cleartext_object == NULL)
     {
@@ -611,11 +632,18 @@ handle_unlock (UDisksEncrypted        *encrypted,
                                         g_udev_device_get_sysfs_attr (cleartext_device->udev_device, "dm/uuid"),
                                         caller_uid);
 
+  /* ensure property changes are sent before the method return */
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (encrypted));
+
   udisks_encrypted_complete_unlock (encrypted,
                                     invocation,
                                     g_dbus_object_get_object_path (G_DBUS_OBJECT (cleartext_object)));
 
  out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_free (device);
   g_free (crypttab_name);
   g_free (crypttab_passphrase);
@@ -656,6 +684,7 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
   gchar *cleartext_path = NULL;
   void *close_func;
   gboolean is_luks;
+  gboolean is_bitlk;
   gboolean handle_as_tcrypt;
 
   object = udisks_daemon_util_dup_object (encrypted, error);
@@ -669,6 +698,7 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
   state = udisks_daemon_get_state (daemon);
   is_luks = udisks_linux_block_is_luks (block);
+  is_bitlk = udisks_linux_block_is_bitlk (block);
   handle_as_tcrypt = udisks_linux_block_is_tcrypt (block) || udisks_linux_block_is_unknown_crypto (block);
 
   /* TODO: check if the device is mentioned in /etc/crypttab (see crypttab(5)) - if so use that
@@ -677,12 +707,12 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
    */
 
   /* Fail if the device is not a LUKS or possible TCRYPT device */
-  if (!(is_luks || handle_as_tcrypt))
+  if (!(is_luks || is_bitlk || handle_as_tcrypt))
     {
       g_set_error (error,
                    UDISKS_ERROR,
                    UDISKS_ERROR_FAILED,
-                   "Device %s does not appear to be a LUKS or TCRYPT device",
+                   "Device %s does not appear to be a LUKS, BITLK or TCRYPT device",
                    udisks_block_get_device (block));
       ret = FALSE;
       goto out;
@@ -757,6 +787,8 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
 
   if (is_luks)
     close_func = luks_close_job_func;
+  else if (is_bitlk)
+    close_func = bitlk_close_job_func;
   else
     close_func = tcrypt_close_job_func;
 
@@ -791,7 +823,7 @@ udisks_linux_encrypted_lock (UDisksLinuxEncrypted   *encrypted,
                                                          wait_for_cleartext_object,
                                                          cleartext_path,
                                                          NULL,
-                                                         10,
+                                                         UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                          &loc_error))
     {
       g_set_error (error,
@@ -827,12 +859,30 @@ handle_lock (UDisksEncrypted        *encrypted,
              GVariant               *options)
 {
   GError *error = NULL;
+  UDisksObject *object = NULL;
+  UDisksDaemon *daemon = NULL;
+  UDisksState *state = NULL;
+
+  object = udisks_daemon_util_dup_object (encrypted, NULL);
+  if (object != NULL)
+    {
+      daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+      state = udisks_daemon_get_state (daemon);
+      udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+      udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+    }
 
   if (!udisks_linux_encrypted_lock (UDISKS_LINUX_ENCRYPTED (encrypted), invocation, options, &error))
     g_dbus_method_invocation_take_error (invocation, error);
   else
     udisks_encrypted_complete_lock (encrypted, invocation);
 
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
+  g_clear_object (&object);
+
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
 
@@ -849,6 +899,7 @@ handle_change_passphrase (UDisksEncrypted        *encrypted,
   UDisksObject *object = NULL;
   UDisksBlock *block;
   UDisksDaemon *daemon;
+  UDisksState *state = NULL;
   uid_t caller_uid;
   const gchar *action_id;
   GError *error = NULL;
@@ -864,6 +915,10 @@ handle_change_passphrase (UDisksEncrypted        *encrypted,
 
   block = udisks_object_peek_block (object);
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
+
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
 
   /* TODO: check if the device is mentioned in /etc/crypttab (see crypttab(5)) - if so use that
    *
@@ -948,6 +1003,10 @@ handle_change_passphrase (UDisksEncrypted        *encrypted,
   udisks_encrypted_complete_change_passphrase (encrypted, invocation);
 
  out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_free (device);
   udisks_string_wipe_and_free (data.passphrase);
   udisks_string_wipe_and_free (data.new_passphrase);
@@ -970,6 +1029,7 @@ handle_resize (UDisksEncrypted       *encrypted,
   UDisksObject *cleartext_object = NULL;
   UDisksBlock *cleartext_block;
   UDisksDaemon *daemon;
+  UDisksState *state = NULL;
   uid_t caller_uid;
   const gchar *action_id = NULL;
   const gchar *message = NULL;
@@ -986,6 +1046,10 @@ handle_resize (UDisksEncrypted       *encrypted,
 
   block = udisks_object_peek_block (object);
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
+
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
 
   /* Fail if the device is not a LUKS device */
   if (!(g_strcmp0 (udisks_block_get_id_usage (block), "crypto") == 0 &&
@@ -1098,6 +1162,10 @@ handle_resize (UDisksEncrypted       *encrypted,
   udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
 
  out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_clear_object (&cleartext_object);
   g_clear_object (&object);
   g_clear_error (&error);
