@@ -32,6 +32,7 @@
 
 #include <limits.h>
 #include <stdlib.h>
+#include <errno.h>
 
 #include "udisksdaemon.h"
 #include "udisksstate.h"
@@ -41,6 +42,7 @@
 #include "udiskslinuxprovider.h"
 #include "udisksdaemonutil.h"
 #include "udiskslinuxencryptedhelpers.h"
+#include "udiskslinuxblockobject.h"
 
 /**
  * SECTION:udisksstate
@@ -77,6 +79,17 @@
  *           <literal>fstab-mount</literal>
  *           (of type <link linkend="G-VARIANT-TYPE-BOOLEAN:CAPS">'b'</link>) that is %TRUE
  *           if the device was mounted via an entry in /etc/fstab.
+ *
+ *           Contains only non-persistent mount points. This state file is typically stored on
+ *           a non-persistent filesystem.
+ *         </entry>
+ *       </row>
+ *       <row>
+ *         <entry><filename>/var/lib/udisks2/mounted-fs-persistent</filename></entry>
+ *         <entry>
+ *           Similar in contents and data structure to <filename>/run/udisks2/mounted-fs</filename>
+ *           except that only persistent mount points are stored there. This state file is
+ *           typically stored on a persistent filesystem.
  *         </entry>
  *       </row>
  *       <row>
@@ -123,6 +136,17 @@
  *           of the user who started the array.
  *         </entry>
  *       </row>
+ *       <row>
+ *         <entry><filename>/run/udisks2/modules</filename></entry>
+ *         <entry>
+ *           A serialized 'a{sa{sv}}' #GVariant mapping from the
+ *           module name (e.g. <filename>lvm2</filename>) into a set of details.
+ *           No details are defined at this point, reserved for future use.
+ *
+ *           Contains currently active module names primarily for the purpose
+ *           of crash recovery upon next daemon start.
+ *         </entry>
+ *       </row>
  *     </tbody>
  *   </tgroup>
  * </table>
@@ -140,6 +164,14 @@
  * to the attention of the system administrator.
  */
 
+/* State file filenames */
+#define UDISKS_STATE_FILE_MOUNTED_FS             "mounted-fs"
+#define UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT  "mounted-fs-persistent"
+#define UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV    "unlocked-crypto-dev"
+#define UDISKS_STATE_FILE_LOOP                   "loop"
+#define UDISKS_STATE_FILE_MDRAID                 "mdraid"
+#define UDISKS_STATE_FILE_MODULES                "modules"
+
 /**
  * UDisksState:
  *
@@ -177,7 +209,9 @@ enum
 
 static void      udisks_state_check_in_thread     (UDisksState          *state);
 static void      udisks_state_check_mounted_fs    (UDisksState          *state,
-                                                   GArray               *devs_to_clean);
+                                                   const gchar          *key,
+                                                   GArray               *devs_to_clean,
+                                                   dev_t                 match_block_device);
 static void      udisks_state_check_unlocked_crypto_dev (UDisksState          *state,
                                                          gboolean              check_only,
                                                          GArray               *devs_to_clean);
@@ -187,10 +221,10 @@ static void      udisks_state_check_loop          (UDisksState          *state,
 static void      udisks_state_check_mdraid        (UDisksState          *state,
                                                    gboolean              check_only,
                                                    GArray               *devs_to_clean);
+static gchar    *get_state_file_path              (const gchar          *key);
 static GVariant *udisks_state_get                 (UDisksState          *state,
                                                    const gchar          *key,
-                                                   const GVariantType   *type,
-                                                   gboolean             *error);
+                                                   const GVariantType   *type);
 static gboolean  udisks_state_set                 (UDisksState          *state,
                                                    const gchar          *key,
                                                    const GVariantType   *type,
@@ -449,6 +483,37 @@ udisks_state_check_sync (UDisksState *state)
   g_mutex_clear (&data.data_mutex);
 }
 
+/**
+ * udisks_state_check_block:
+ * @state: A #UDisksState.
+ * @block_device: Device number of the block device to check.
+ *
+ * Performs cleanup of a mounted filesystem over a single block device. In case
+ * the clean-up thread is busy, waits until the current iteration is finished.
+ *
+ * This can be called from any thread and will block the calling thread until
+ * cleanup is finished. Note that this ignores #UDisksLinuxBlockObject cleanup lock
+ * that is supposed to be held by the caller.
+ */
+void
+udisks_state_check_block (UDisksState   *state,
+                          dev_t          block_device)
+{
+
+  g_mutex_lock (&state->lock);
+
+  udisks_state_check_mounted_fs (state,
+                                 UDISKS_STATE_FILE_MOUNTED_FS,
+                                 NULL,
+                                 block_device);
+  udisks_state_check_mounted_fs (state,
+                                 UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT,
+                                 NULL,
+                                 block_device);
+
+  g_mutex_unlock (&state->lock);
+}
+
 /**
  * udisks_state_get_daemon:
  * @state: A #UDisksState.
@@ -498,7 +563,14 @@ udisks_state_check_in_thread (UDisksState *state)
   /* Then go through all mounted filesystems and pass the
    * devices that we intend to clean...
    */
-  udisks_state_check_mounted_fs (state, devs_to_clean);
+  udisks_state_check_mounted_fs (state,
+                                 UDISKS_STATE_FILE_MOUNTED_FS,
+                                 devs_to_clean,
+                                 0);
+  udisks_state_check_mounted_fs (state,
+                                 UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT,
+                                 devs_to_clean,
+                                 0);
 
   /* Then go through all block devices and clear them up
    * ... for real this time
@@ -514,7 +586,7 @@ udisks_state_check_in_thread (UDisksState *state)
                              FALSE, /* check_only */
                              NULL);
 
-  g_array_unref (devs_to_clean);
+  g_array_free (devs_to_clean, TRUE);
 
   udisks_info ("Cleanup check end");
 
@@ -587,7 +659,8 @@ trigger_change_uevent (const gchar *sysfs_path)
 static gboolean
 udisks_state_check_mounted_fs_entry (UDisksState  *state,
                                      GVariant     *value,
-                                     GArray       *devs_to_clean)
+                                     GArray       *devs_to_clean,
+                                     dev_t         match_block_device)
 {
   const gchar *mount_point_str;
   gchar mount_point[PATH_MAX] = { '\0', };
@@ -603,54 +676,77 @@ udisks_state_check_mounted_fs_entry (UDisksState  *state,
   gboolean is_mounted;
   gboolean device_exists;
   gboolean device_to_be_cleaned;
-  gboolean attempt_no_cleanup;
   UDisksMountMonitor *monitor;
   GUdevClient *udev_client;
   GUdevDevice *udev_device;
   guint n;
   gchar *change_sysfs_path = NULL;
+  UDisksObject *block_object = NULL;
+  gboolean locked;
 
   keep = FALSE;
   is_mounted = FALSE;
   device_exists = FALSE;
   device_to_be_cleaned = FALSE;
-  attempt_no_cleanup = FALSE;
   block_device_value = NULL;
   fstab_mount_value = NULL;
   fstab_mount = FALSE;
   details = NULL;
+  locked = FALSE;
 
   monitor = udisks_daemon_get_mount_monitor (state->daemon);
-  udev_client = udisks_linux_provider_get_udev_client (udisks_daemon_get_linux_provider (state->daemon));
 
   g_variant_get (value,
                  "{&s@a{sv}}",
                  &mount_point_str,
                  &details);
 
-  if (realpath (mount_point_str, mount_point) == NULL)
-    {
-      udisks_critical ("mountpoint %s is invalid, cannot recover the canonical path ", mount_point_str);
-    }
-
   block_device_value = lookup_asv (details, "block-device");
   if (block_device_value == NULL)
     {
       s = g_variant_print (value, TRUE);
-      udisks_critical ("mounted-fs entry %s is invalid: no block-device key/value pair", s);
+      udisks_critical ("udisks_state_check_mounted_fs_entry: mounted-fs entry %s is invalid: no block-device key/value pair", s);
       g_free (s);
-      attempt_no_cleanup = FALSE;
       goto out;
     }
   block_device = g_variant_get_uint64 (block_device_value);
 
+  if (match_block_device != 0 && match_block_device != block_device)
+    {
+      /* not intended for this block device, continue with parent iteration */
+      keep = TRUE;
+      goto out;
+    }
+
+  block_object = udisks_daemon_find_block (state->daemon, block_device);
+  /* skip locking if called from udisks_state_check_block() */
+  if (block_object != NULL && match_block_device == 0)
+    {
+      if (! udisks_linux_block_object_try_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (block_object)))
+        {
+          gchar *device_file;
+
+          device_file = udisks_linux_block_object_get_device_file (UDISKS_LINUX_BLOCK_OBJECT (block_object));
+          udisks_notice ("udisks_state_check_mounted_fs_entry: block device %s is busy, skipping cleanup", device_file);
+          g_free (device_file);
+
+          keep = TRUE;
+          goto out;
+        }
+      locked = TRUE;
+    }
+
+  if (realpath (mount_point_str, mount_point) == NULL)
+    {
+      udisks_critical ("udisks_state_check_mounted_fs_entry: mountpoint %s is invalid, cannot recover the canonical path ", mount_point_str);
+    }
+
   fstab_mount_value = lookup_asv (details, "fstab-mount");
   if (fstab_mount_value == NULL)
     {
       s = g_variant_print (value, TRUE);
-      udisks_critical ("mounted-fs entry %s is invalid: no fstab-mount key/value pair", s);
+      udisks_critical ("udisks_state_check_mounted_fs_entry: mounted-fs entry %s is invalid: no fstab-mount key/value pair", s);
       g_free (s);
-      attempt_no_cleanup = FALSE;
       goto out;
     }
   fstab_mount = g_variant_get_boolean (fstab_mount_value);
@@ -672,6 +768,7 @@ udisks_state_check_mounted_fs_entry (UDisksState  *state,
   g_list_free_full (mounts, g_object_unref);
 
   /* Figure out if block device still exists */
+  udev_client = udisks_linux_provider_get_udev_client (udisks_daemon_get_linux_provider (state->daemon));
   udev_device = g_udev_client_query_by_device_number (udev_client,
                                                       G_UDEV_DEVICE_TYPE_BLOCK,
                                                       block_device);
@@ -727,22 +824,23 @@ udisks_state_check_mounted_fs_entry (UDisksState  *state,
     }
 
   /* Figure out if the device is about to be cleaned up */
-  for (n = 0; n < devs_to_clean->len; n++)
-    {
-      dev_t dev_to_clean = g_array_index (devs_to_clean, dev_t, n);
-      if (dev_to_clean == block_device)
-        {
-          device_to_be_cleaned = TRUE;
-          break;
-        }
-    }
+  if (devs_to_clean != NULL)
+    for (n = 0; n < devs_to_clean->len; n++)
+      {
+        dev_t dev_to_clean = g_array_index (devs_to_clean, dev_t, n);
+        if (dev_to_clean == block_device)
+          {
+            device_to_be_cleaned = TRUE;
+            break;
+          }
+      }
 
   if (is_mounted && device_exists && !device_to_be_cleaned)
     keep = TRUE;
 
  out:
 
-  if (!keep && !attempt_no_cleanup)
+  if (!keep)
     {
       if (!device_exists)
         {
@@ -824,6 +922,9 @@ udisks_state_check_mounted_fs_entry (UDisksState  *state,
     g_variant_unref (block_device_value);
   if (details != NULL)
     g_variant_unref (details);
+  if (locked)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (block_object));
+  g_clear_object (&block_object);
 
   g_free (change_sysfs_path);
 
@@ -833,23 +934,21 @@ udisks_state_check_mounted_fs_entry (UDisksState  *state,
 /* called with mutex->lock held */
 static void
 udisks_state_check_mounted_fs (UDisksState *state,
-                               GArray      *devs_to_clean)
+                               const gchar *key,
+                               GArray      *devs_to_clean,
+                               dev_t        match_block_device)
 {
   gboolean changed;
   GVariant *value;
   GVariant *new_value;
   GVariantBuilder builder;
 
-  gboolean ok = FALSE;
-
   changed = FALSE;
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mounted-fs",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            key,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
 
   /* check valid entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sa{sv}}"));
@@ -860,7 +959,7 @@ udisks_state_check_mounted_fs (UDisksState *state,
       g_variant_iter_init (&iter, value);
       while ((child = g_variant_iter_next_value (&iter)) != NULL)
         {
-          if (udisks_state_check_mounted_fs_entry (state, child, devs_to_clean))
+          if (udisks_state_check_mounted_fs_entry (state, child, devs_to_clean, match_block_device))
             g_variant_builder_add_value (&builder, child);
           else
             changed = TRUE;
@@ -875,7 +974,7 @@ udisks_state_check_mounted_fs (UDisksState *state,
   if (changed)
     {
       udisks_state_set (state,
-                        "mounted-fs",
+                        key,
                         G_VARIANT_TYPE ("a{sa{sv}}"),
                         new_value /* consumes new_value */);
     }
@@ -883,9 +982,6 @@ udisks_state_check_mounted_fs (UDisksState *state,
     {
       g_variant_unref (new_value);
     }
-
- out:
-  ;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -897,6 +993,7 @@ udisks_state_check_mounted_fs (UDisksState *state,
  * @mount_point: The mount point.
  * @uid: The user id of the process requesting the device to be mounted.
  * @fstab_mount: %TRUE if the device was mounted via /etc/fstab.
+ * @persistent: %TRUE if the mount point is on a persistent filesystem.
  *
  * Adds a new entry to the
  * <filename>/run/udisks2/mounted-fs</filename> file.
@@ -906,7 +1003,8 @@ udisks_state_add_mounted_fs (UDisksState    *state,
                              const gchar    *mount_point,
                              dev_t           block_device,
                              uid_t           uid,
-                             gboolean        fstab_mount)
+                             gboolean        fstab_mount,
+                             gboolean        persistent)
 {
   GVariant *value;
   GVariant *new_value;
@@ -914,8 +1012,6 @@ udisks_state_add_mounted_fs (UDisksState    *state,
   GVariantBuilder builder;
   GVariantBuilder details_builder;
 
-  gboolean ok = FALSE;
-
   g_return_if_fail (UDISKS_IS_STATE (state));
   g_return_if_fail (mount_point != NULL);
 
@@ -923,10 +1019,8 @@ udisks_state_add_mounted_fs (UDisksState    *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mounted-fs",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            persistent ? UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT : UDISKS_STATE_FILE_MOUNTED_FS,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
 
   /* start by including existing entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sa{sv}}"));
@@ -980,49 +1074,28 @@ udisks_state_add_mounted_fs (UDisksState    *state,
 
   /* save new entries */
   udisks_state_set (state,
-                         "mounted-fs",
-                         G_VARIANT_TYPE ("a{sa{sv}}"),
-                         new_value /* consumes new_value */);
+                    persistent ? UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT : UDISKS_STATE_FILE_MOUNTED_FS,
+                    G_VARIANT_TYPE ("a{sa{sv}}"),
+                    new_value /* consumes new_value */);
 
- out:
   g_mutex_unlock (&state->lock);
 }
 
-/**
- * udisks_state_find_mounted_fs:
- * @state: A #UDisksState.
- * @block_device: The block device.
- * @out_uid: Return location for the user id who mounted the device or %NULL.
- * @out_fstab_mount: Return location for whether the device was a fstab mount or %NULL.
- *
- * Gets the mount point for @block_device, if it exists in the
- * <filename>/run/udisks2/mounted-fs</filename> file.
- *
- * Returns: The mount point for @block_device or %NULL if not found.
- */
-gchar *
-udisks_state_find_mounted_fs (UDisksState   *state,
-                              dev_t          block_device,
-                              uid_t         *out_uid,
-                              gboolean      *out_fstab_mount)
+/* called with state->lock held */
+static gchar *
+find_mounted_fs_for_key (UDisksState   *state,
+                         const gchar   *key,
+                         dev_t          block_device,
+                         uid_t         *out_uid,
+                         gboolean      *out_fstab_mount)
 {
-  gchar *ret;
+  gchar *ret = NULL;
   GVariant *value;
-  gboolean ok = FALSE;
-
-  g_return_val_if_fail (UDISKS_IS_STATE (state), NULL);
-
-  g_mutex_lock (&state->lock);
-
-  ret = NULL;
-  value = NULL;
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mounted-fs",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            key,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
 
   /* look through list */
   if (value != NULL)
@@ -1086,10 +1159,49 @@ udisks_state_find_mounted_fs (UDisksState   *state,
  out:
   if (value != NULL)
     g_variant_unref (value);
-  g_mutex_unlock (&state->lock);
   return ret;
 }
 
+/**
+ * udisks_state_find_mounted_fs:
+ * @state: A #UDisksState.
+ * @block_device: The block device.
+ * @out_uid: Return location for the user id who mounted the device or %NULL.
+ * @out_fstab_mount: Return location for whether the device was a fstab mount or %NULL.
+ *
+ * Gets the mount point for @block_device, if it exists in the
+ * <filename>/run/udisks2/mounted-fs</filename> file.
+ *
+ * Returns: The mount point for @block_device or %NULL if not found.
+ */
+gchar *
+udisks_state_find_mounted_fs (UDisksState   *state,
+                              dev_t          block_device,
+                              uid_t         *out_uid,
+                              gboolean      *out_fstab_mount)
+{
+  gchar *ret;
+
+  g_return_val_if_fail (UDISKS_IS_STATE (state), NULL);
+
+  g_mutex_lock (&state->lock);
+
+  ret = find_mounted_fs_for_key (state,
+                                 UDISKS_STATE_FILE_MOUNTED_FS,
+                                 block_device,
+                                 out_uid,
+                                 out_fstab_mount);
+  if (ret == NULL)
+    ret = find_mounted_fs_for_key (state,
+                                   UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT,
+                                   block_device,
+                                   out_uid,
+                                   out_fstab_mount);
+
+  g_mutex_unlock (&state->lock);
+
+  return ret;
+}
 /* ---------------------------------------------------------------------------------------------------- */
 
 /* returns TRUE if the entry should be kept */
@@ -1264,16 +1376,12 @@ udisks_state_check_unlocked_crypto_dev (UDisksState *state,
   GVariant *new_value;
   GVariantBuilder builder;
 
-  gboolean ok = FALSE;
-
   changed = FALSE;
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "unlocked-crypto-dev",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
 
   /* check valid entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{ta{sv}}"));
@@ -1299,7 +1407,7 @@ udisks_state_check_unlocked_crypto_dev (UDisksState *state,
   if (changed)
     {
       udisks_state_set (state,
-                        "unlocked-crypto-dev",
+                        UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV,
                         G_VARIANT_TYPE ("a{ta{sv}}"),
                         new_value /* consumes new_value */);
     }
@@ -1307,9 +1415,6 @@ udisks_state_check_unlocked_crypto_dev (UDisksState *state,
     {
       g_variant_unref (new_value);
     }
-
- out:
-  ;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1338,8 +1443,6 @@ udisks_state_add_unlocked_crypto_dev (UDisksState  *state,
   GVariantBuilder builder;
   GVariantBuilder details_builder;
 
-  gboolean ok = FALSE;
-
   g_return_if_fail (UDISKS_IS_STATE (state));
   g_return_if_fail (dm_uuid != NULL);
 
@@ -1347,10 +1450,8 @@ udisks_state_add_unlocked_crypto_dev (UDisksState  *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "unlocked-crypto-dev",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
 
   /* start by including existing entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{ta{sv}}"));
@@ -1404,10 +1505,10 @@ udisks_state_add_unlocked_crypto_dev (UDisksState  *state,
 
   /* save new entries */
   udisks_state_set (state,
-                         "unlocked-crypto-dev",
-                         G_VARIANT_TYPE ("a{ta{sv}}"),
-                         new_value /* consumes new_value */);
- out:
+                    UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV,
+                    G_VARIANT_TYPE ("a{ta{sv}}"),
+                    new_value /* consumes new_value */);
+
   g_mutex_unlock (&state->lock);
 }
 
@@ -1429,7 +1530,6 @@ udisks_state_find_unlocked_crypto_dev (UDisksState   *state,
 {
   dev_t ret;
   GVariant *value;
-  gboolean ok = FALSE;
 
   g_return_val_if_fail (UDISKS_IS_STATE (state), 0);
 
@@ -1440,10 +1540,8 @@ udisks_state_find_unlocked_crypto_dev (UDisksState   *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "unlocked-crypto-dev",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_UNLOCKED_CRYPTO_DEV,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
 
   /* look through list */
   if (value != NULL)
@@ -1530,7 +1628,7 @@ udisks_state_check_loop_entry (UDisksState  *state,
     {
       gchar *s;
       s = g_variant_print (value, TRUE);
-      udisks_critical ("loop entry %s is invalid: no backing-file key/value pair", s);
+      udisks_critical ("udisks_state_check_loop_entry: loop entry %s is invalid: no backing-file key/value pair", s);
       g_free (s);
       goto out;
     }
@@ -1540,12 +1638,12 @@ udisks_state_check_loop_entry (UDisksState  *state,
   device = g_udev_client_query_by_device_file (udev_client, loop_device);
   if (device == NULL)
     {
-      udisks_info ("no udev device for %s", loop_device);
+      udisks_info ("udisks_state_check_loop_entry: no udev device for %s", loop_device);
       goto out;
     }
   if (g_udev_device_get_sysfs_attr (device, "loop/offset") == NULL)
     {
-      udisks_info ("loop device %s is not setup  (no loop/offset sysfs file)", loop_device);
+      udisks_info ("udisks_state_check_loop_entry: loop device %s is not setup  (no loop/offset sysfs file)", loop_device);
       goto out;
     }
 
@@ -1560,7 +1658,7 @@ udisks_state_check_loop_entry (UDisksState  *state,
   sysfs_backing_file = g_udev_device_get_sysfs_attr (device, "loop/backing_file");
   if (g_strcmp0 (sysfs_backing_file, backing_file) != 0)
     {
-      udisks_notice ("unexpected name for %s - expected `%s' but got `%s'",
+      udisks_notice ("udisks_state_check_loop_entry: unexpected name for %s - expected `%s' but got `%s'",
                      loop_device, backing_file, sysfs_backing_file);
       goto out;
     }
@@ -1605,16 +1703,12 @@ udisks_state_check_loop (UDisksState *state,
   GVariant *new_value;
   GVariantBuilder builder;
 
-  gboolean ok = FALSE;
-
   changed = FALSE;
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "loop",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_LOOP,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
 
   /* check valid entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sa{sv}}"));
@@ -1640,7 +1734,7 @@ udisks_state_check_loop (UDisksState *state,
   if (changed)
     {
       udisks_state_set (state,
-                        "loop",
+                        UDISKS_STATE_FILE_LOOP,
                         G_VARIANT_TYPE ("a{sa{sv}}"),
                         new_value /* consumes new_value */);
     }
@@ -1648,9 +1742,6 @@ udisks_state_check_loop (UDisksState *state,
     {
       g_variant_unref (new_value);
     }
-
- out:
-  ;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1679,8 +1770,6 @@ udisks_state_add_loop (UDisksState   *state,
   GVariantBuilder builder;
   GVariantBuilder details_builder;
 
-  gboolean ok = FALSE;
-
   g_return_if_fail (UDISKS_IS_STATE (state));
   g_return_if_fail (device_file != NULL);
   g_return_if_fail (backing_file != NULL);
@@ -1689,10 +1778,8 @@ udisks_state_add_loop (UDisksState   *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "loop",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_LOOP,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
 
   /* start by including existing entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sa{sv}}"));
@@ -1745,10 +1832,10 @@ udisks_state_add_loop (UDisksState   *state,
 
   /* save new entries */
   udisks_state_set (state,
-                    "loop",
+                    UDISKS_STATE_FILE_LOOP,
                     G_VARIANT_TYPE ("a{sa{sv}}"),
                     new_value /* consumes new_value */);
- out:
+
   g_mutex_unlock (&state->lock);
 }
 
@@ -1831,7 +1918,6 @@ _udisks_state_has_loop_list_visitor (GVariant *child, gpointer compare_data,
  * @state: A #UDisksState
  * @device_file: A loop device file.
  * @out_uid: Return location for the user id who setup the loop device or %NULL.
- * @error: Return location for error or %NULL.
  *
  * Checks if @device_file is set up via udisks.
  *
@@ -1844,7 +1930,6 @@ udisks_state_has_loop (UDisksState   *state,
 {
   gboolean ret;
   GVariant *value;
-  gboolean ok = FALSE;
 
   g_return_val_if_fail (UDISKS_IS_STATE (state), FALSE);
 
@@ -1855,9 +1940,9 @@ udisks_state_has_loop (UDisksState   *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "loop",
-                            G_VARIANT_TYPE ("a{sa{sv}}"), &ok);
-  if (ok && value)
+                            UDISKS_STATE_FILE_LOOP,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
+  if (value != NULL)
    {
       ret = iterate_list (value,
                           _udisks_state_has_loop_list_visitor,
@@ -1896,13 +1981,13 @@ udisks_state_check_mdraid_entry (UDisksState  *state,
   device = g_udev_client_query_by_device_number (udev_client, G_UDEV_DEVICE_TYPE_BLOCK, raid_device);
   if (device == NULL)
     {
-      udisks_info ("no udev device for raid device %u:%u", major (raid_device), minor (raid_device));
+      udisks_info ("udisks_state_check_mdraid_entry: no udev device for raid device %u:%u", major (raid_device), minor (raid_device));
       goto out;
     }
   array_state = g_udev_device_get_sysfs_attr (device, "md/array_state");
   if (array_state == NULL)
     {
-      udisks_info ("raid device %u:%u is not setup  (no md/array_state sysfs file)",
+      udisks_info ("udisks_state_check_mdraid_entry: raid device %u:%u is not setup  (no md/array_state sysfs file)",
                    major (raid_device), minor (raid_device));
       goto out;
     }
@@ -1950,16 +2035,12 @@ udisks_state_check_mdraid (UDisksState *state,
   GVariant *new_value;
   GVariantBuilder builder;
 
-  gboolean ok = FALSE;
-
   changed = FALSE;
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mdraid",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_MDRAID,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
 
   /* check valid entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{ta{sv}}"));
@@ -1985,7 +2066,7 @@ udisks_state_check_mdraid (UDisksState *state,
   if (changed)
     {
       udisks_state_set (state,
-                        "mdraid",
+                        UDISKS_STATE_FILE_MDRAID,
                         G_VARIANT_TYPE ("a{ta{sv}}"),
                         new_value /* consumes new_value */);
     }
@@ -1993,9 +2074,6 @@ udisks_state_check_mdraid (UDisksState *state,
     {
       g_variant_unref (new_value);
     }
-
- out:
-  ;
 }
 
 /**
@@ -2017,7 +2095,6 @@ udisks_state_add_mdraid (UDisksState   *state,
   GVariant *details_value;
   GVariantBuilder builder;
   GVariantBuilder details_builder;
-  gboolean ok = FALSE;
 
   g_return_if_fail (UDISKS_IS_STATE (state));
 
@@ -2025,10 +2102,8 @@ udisks_state_add_mdraid (UDisksState   *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mdraid",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (!ok)
-    goto out;
+                            UDISKS_STATE_FILE_MDRAID,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
 
   /* start by including existing entries */
   g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{ta{sv}}"));
@@ -2073,11 +2148,10 @@ udisks_state_add_mdraid (UDisksState   *state,
 
   /* save new entries */
   udisks_state_set (state,
-                         "mdraid",
-                         G_VARIANT_TYPE ("a{ta{sv}}"),
-                         new_value /* consumes new_value */);
+                    UDISKS_STATE_FILE_MDRAID,
+                    G_VARIANT_TYPE ("a{ta{sv}}"),
+                    new_value /* consumes new_value */);
 
- out:
   g_mutex_unlock (&state->lock);
 }
 
@@ -2117,7 +2191,6 @@ _udisks_state_has_mdraid_list_visitor (GVariant *child, gpointer compare_data,
  * @state: A #UDisksState
  * @raid_device: A #dev_t for the RAID device.
  * @out_uid: Return location for the user id who setup the loop device or %NULL.
- * @error: Return location for error or %NULL.
  *
  * Checks if @raid_device is set up via udisks.
  *
@@ -2130,7 +2203,6 @@ udisks_state_has_mdraid (UDisksState   *state,
 {
   gboolean ret = FALSE;
   GVariant *value = NULL;
-  gboolean ok = FALSE;
 
   g_return_val_if_fail (UDISKS_IS_STATE (state), FALSE);
 
@@ -2138,9 +2210,9 @@ udisks_state_has_mdraid (UDisksState   *state,
 
   /* load existing entries */
   value = udisks_state_get (state,
-                            "mdraid",
-                            G_VARIANT_TYPE ("a{ta{sv}}"), &ok);
-  if (ok && value)
+                            UDISKS_STATE_FILE_MDRAID,
+                            G_VARIANT_TYPE ("a{ta{sv}}"));
+  if (value != NULL)
     {
       ret = iterate_list (value, _udisks_state_has_mdraid_list_visitor,
                           (gpointer) &raid_device, (gpointer) out_uid);
@@ -2153,24 +2225,175 @@ udisks_state_has_mdraid (UDisksState   *state,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+/**
+ * udisks_state_add_module:
+ * @state: A #UDisksState.
+ * @module_name: The #UDisksModule name.
+ *
+ * Adds a new entry to the <filename>/run/udisks2/modules</filename>
+ * file. The @module_name string should be the one returned by
+ * udisks_module_get_name().
+ */
+void
+udisks_state_add_module (UDisksState *state,
+                         const gchar *module_name)
+{
+  GVariant *value;
+  GVariant *new_value;
+  GVariantBuilder builder;
+
+  g_return_if_fail (UDISKS_IS_STATE (state));
+
+  g_mutex_lock (&state->lock);
+
+  /* load existing entries */
+  value = udisks_state_get (state,
+                            UDISKS_STATE_FILE_MODULES,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
+
+  /* start by including existing entries */
+  g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{sa{sv}}"));
+  if (value != NULL)
+    {
+      GVariantIter iter;
+      GVariant *child;
+      g_variant_iter_init (&iter, value);
+      while ((child = g_variant_iter_next_value (&iter)) != NULL)
+        {
+          const gchar *entry_module_name;
+          g_variant_get (child, "{&s@a{sv}}", &entry_module_name, NULL);
+          /* Skip/remove stale entries */
+          if (g_strcmp0 (entry_module_name, module_name) == 0)
+            {
+              udisks_warning ("Removing stale entry for module '%s' in /run/udisks2/modules file",
+                              entry_module_name);
+            }
+          else
+            {
+              g_variant_builder_add_value (&builder, child);
+            }
+          g_variant_unref (child);
+        }
+      g_variant_unref (value);
+    }
+
+  /* finally add the new entry */
+  g_variant_builder_add (&builder,
+                         "{s@a{sv}}",
+                         module_name,
+                         g_variant_new_array (G_VARIANT_TYPE ("{sv}"), NULL, 0));
+  new_value = g_variant_builder_end (&builder);
+
+  /* save new entries */
+  udisks_state_set (state,
+                    UDISKS_STATE_FILE_MODULES,
+                    G_VARIANT_TYPE ("a{sa{sv}}"),
+                    new_value /* consumes new_value */);
+
+  g_mutex_unlock (&state->lock);
+}
+
+/**
+ * udisks_state_clear_modules:
+ * @state: A #UDisksState.
+ *
+ * Removes all entries from the <filename>/run/udisks2/modules</filename>
+ * state file.
+ */
+void
+udisks_state_clear_modules (UDisksState *state)
+{
+  gchar *path;
+
+  g_return_if_fail (UDISKS_IS_STATE (state));
+
+  g_mutex_lock (&state->lock);
+
+  /* just remove the file entirely */
+  path = get_state_file_path (UDISKS_STATE_FILE_MODULES);
+  if (g_unlink (path))
+    {
+      if (errno != ENOENT)
+        g_warning ("Error removing state file %s: %m", path);
+    }
+  g_free (path);
+
+  g_mutex_unlock (&state->lock);
+}
+
+/**
+ * udisks_state_get_modules:
+ * @state: A #UDisksState
+ *
+ * Retrieves list of modules recorded in the <filename>/run/udisks2/modules</filename>
+ * state file.
+ *
+ * Returns: (transfer full) (array zero-terminated=1): A list of module names. Free with g_strfreev().
+ */
+gchar **
+udisks_state_get_modules (UDisksState *state)
+{
+  GPtrArray *list;
+  GVariant *value;
+
+  g_return_val_if_fail (UDISKS_IS_STATE (state), NULL);
+
+  g_mutex_lock (&state->lock);
+
+  list = g_ptr_array_new ();
+
+  value = udisks_state_get (state,
+                            UDISKS_STATE_FILE_MODULES,
+                            G_VARIANT_TYPE ("a{sa{sv}}"));
+  if (value != NULL)
+    {
+      GVariantIter iter;
+      GVariant *child;
+      g_variant_iter_init (&iter, value);
+      while ((child = g_variant_iter_next_value (&iter)) != NULL)
+        {
+          gchar *entry_module_name;
+          g_variant_get (child, "{s@a{sv}}", &entry_module_name, NULL);
+          g_ptr_array_add (list, entry_module_name);
+          g_variant_unref (child);
+        }
+      g_variant_unref (value);
+    }
+
+  g_mutex_unlock (&state->lock);
+
+  g_ptr_array_add (list, NULL);
+
+  return (gchar **) g_ptr_array_free (list, FALSE);
+}
+
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+/* returns fully allocated string */
+static gchar *
+get_state_file_path (const gchar *key)
+{
+  if (g_str_equal (key, UDISKS_STATE_FILE_MOUNTED_FS_PERSISTENT))
+    return g_strdup_printf (PACKAGE_LOCALSTATE_DIR "/lib/udisks2/%s", key);
+
+  return g_strdup_printf ("/run/udisks2/%s", key);
+}
+
 static GVariant *
 udisks_state_get (UDisksState           *state,
                   const gchar           *key,
-                  const GVariantType    *type,
-                  gboolean              *ok)
+                  const GVariantType    *type)
 {
-  gchar *path = NULL;
-  GVariant *ret = NULL;
+  gchar *path;
+  GVariant *ret;
   gchar *contents = NULL;
   GError *local_error = NULL;
   gsize length = 0;
 
-  *ok = TRUE;
-
   g_return_val_if_fail (UDISKS_IS_STATE (state), NULL);
   g_return_val_if_fail (key != NULL, NULL);
   g_return_val_if_fail (g_variant_type_is_definite (type), NULL);
-  g_return_val_if_fail (ok != NULL, NULL);
 
   /* TODO:
    *
@@ -2178,14 +2401,7 @@ udisks_state_get (UDisksState           *state,
    * - could also mmap the file
    */
 
-#ifdef HAVE_FHS_MEDIA
-  /* /media usually isn't on a tmpfs, so we need to make this persistant */
-  if (strcmp (key, "mounted-fs") == 0)
-    path = g_strdup_printf (PACKAGE_LOCALSTATE_DIR "/lib/udisks2/%s", key);
-  else
-#endif
-    path = g_strdup_printf ("/run/udisks2/%s", key);
-
+  path = get_state_file_path (key);
 
   /* see if it's already in the cache */
   ret = g_hash_table_lookup (state->cache, path);
@@ -2207,8 +2423,7 @@ udisks_state_get (UDisksState           *state,
           goto out;
         }
 
-      *ok = FALSE;
-      udisks_warning ("Error getting %s: %s (%s, %d)",
+      udisks_warning ("Error getting state data %s: %s (%s, %d)",
                       key,
                       local_error->message,
                       g_quark_to_string (local_error->domain),
@@ -2223,6 +2438,7 @@ udisks_state_get (UDisksState           *state,
                                  FALSE,
                                  g_free,
                                  contents);
+  g_warn_if_fail (ret != NULL);
   g_variant_ref_sink (ret);
 
   contents = NULL; /* ownership transfered to the returned GVariant */
@@ -2240,10 +2456,10 @@ udisks_state_set (UDisksState          *state,
                   GVariant             *value)
 {
   gboolean ret = FALSE;
-  gsize size = 0;
-  gchar *path = NULL;
-  gchar *data= NULL;
-  GVariant *normalized = NULL;
+  gsize size;
+  gchar *path;
+  gchar *data;
+  GVariant *normalized;
   GError *error = NULL;
 
   g_return_val_if_fail (UDISKS_IS_STATE (state), FALSE);
@@ -2257,14 +2473,7 @@ udisks_state_set (UDisksState          *state,
   data = g_malloc (size);
   g_variant_store (normalized, data);
 
-#ifdef HAVE_FHS_MEDIA
-  /* /media usually isn't on a tmpfs, so we need to make this persistant */
-  if (strcmp (key, "mounted-fs") == 0)
-    path = g_strdup_printf (PACKAGE_LOCALSTATE_DIR "/lib/udisks2/%s", key);
-  else
-#endif
-    path = g_strdup_printf ("/run/udisks2/%s", key);
-
+  path = get_state_file_path (key);
   g_hash_table_insert (state->cache, g_strdup (path), g_variant_ref (value));
 
   if (!g_file_set_contents (path,
@@ -2272,7 +2481,7 @@ udisks_state_set (UDisksState          *state,
                             size,
                             &error))
     {
-      udisks_warning ("Error setting %s: %s (%s, %d)", key,
+      udisks_warning ("Error setting state data %s: %s (%s, %d)", key,
                      error->message,
                      g_quark_to_string (error->domain),
                      error->code);
